package main.util;

import java.util.Calendar;

/*
 * @author Pankratz
 * Date: 22.03.2006
 * Time: 13:08:28
 */

/**
 * $Id: WufiCalendar.java,v 1.2 2008/04/09 10:34:45 fpankratz Exp $
 */
public class WufiCalendar extends Calendar {
	static final int MONTH_LENGTH[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31,
			30, 31 };

	/**
	 * Value of the <code>ERA</code> field indicating the period before the
	 * common era (before Christ), also known as BCE. The sequence of years at
	 * the transition from <code>BC</code> to <code>AD</code> is ..., 2 BC,
	 * 1 BC, 1 AD, 2 AD,...
	 * 
	 * @see #ERA
	 */
	public static final int BC = 0;

	/**
	 * Value of the {@link #ERA} field indicating the period before the common
	 * era, the same value as {@link #BC}.
	 * 
	 * @see #CE
	 */
	static final int BCE = 0;

	/**
	 * Value of the <code>ERA</code> field indicating the common era (Anno
	 * Domini), also known as CE. The sequence of years at the transition from
	 * <code>BC</code> to <code>AD</code> is ..., 2 BC, 1 BC, 1 AD, 2 AD,...
	 * 
	 * @see #ERA
	 */
	public static final int AD = 1;

	/**
	 * Value of the {@link #ERA} field indicating the common era, the same value
	 * as {@link #AD}.
	 * 
	 * @see #BCE
	 */
	static final int CE = 1;

	private static final int EPOCH_OFFSET = 719163; // Fixed date of January 1,
													// 1970 (Gregorian)
	private static final int EPOCH_YEAR = 1970;

	// Useful millisecond constants. Although ONE_DAY and ONE_WEEK can fit
	// into ints, they must be longs in order to prevent arithmetic overflow
	// when performing (bug 4173516).
	private static final int ONE_SECOND = 1000;
	private static final int ONE_MINUTE = 60 * ONE_SECOND;
	private static final int ONE_HOUR = 60 * ONE_MINUTE;
	private static final long ONE_DAY = 24 * ONE_HOUR;
	private static final long ONE_WEEK = 7 * ONE_DAY;

	/*
	 * <pre> Greatest Least Field name Minimum Minimum Maximum Maximum
	 * ---------- ------- ------- ------- ------- ERA 0 0 1 1 YEAR 1 1 292269054
	 * 292278994 MONTH 0 0 11 11 WEEK_OF_YEAR 1 1 52* 53 WEEK_OF_MONTH 0 0 4* 6
	 * DAY_OF_MONTH 1 1 28* 31 DAY_OF_YEAR 1 1 365* 366 DAY_OF_WEEK 1 1 7 7
	 * DAY_OF_WEEK_IN_MONTH -1 -1 4* 6 AM_PM 0 0 1 1 HOUR 0 0 11 11 HOUR_OF_DAY
	 * 0 0 23 23 MINUTE 0 0 59 59 SECOND 0 0 59 59 MILLISECOND 0 0 999 999
	 * ZONE_OFFSET -13:00 -13:00 14:00 14:00 DST_OFFSET 0:00 0:00 0:20 2:00
	 * </pre> *: depends on the Gregorian change date
	 */
	static final int MIN_VALUES[] = { BCE, // ERA
			1, // YEAR
			JANUARY, // MONTH
			1, // WEEK_OF_YEAR
			0, // WEEK_OF_MONTH
			1, // DAY_OF_MONTH
			1, // DAY_OF_YEAR
			SUNDAY, // DAY_OF_WEEK
			1, // DAY_OF_WEEK_IN_MONTH
			AM, // AM_PM
			0, // HOUR
			0, // HOUR_OF_DAY
			0, // MINUTE
			0, // SECOND
			0, // MILLISECOND
			-13 * ONE_HOUR, // ZONE_OFFSET (UNIX compatibility)
			0 // DST_OFFSET
	};
	static final int LEAST_MAX_VALUES[] = { CE, // ERA
			292269054, // YEAR
			DECEMBER, // MONTH
			52, // WEEK_OF_YEAR
			4, // WEEK_OF_MONTH
			28, // DAY_OF_MONTH
			365, // DAY_OF_YEAR
			SATURDAY, // DAY_OF_WEEK
			4, // DAY_OF_WEEK_IN
			PM, // AM_PM
			11, // HOUR
			23, // HOUR_OF_DAY
			59, // MINUTE
			59, // SECOND
			999, // MILLISECOND
			14 * ONE_HOUR, // ZONE_OFFSET
			20 * ONE_MINUTE // DST_OFFSET (historical least maximum)
	};
	static final int MAX_VALUES[] = { CE, // ERA
			292278994, // YEAR
			DECEMBER, // MONTH
			53, // WEEK_OF_YEAR
			6, // WEEK_OF_MONTH
			31, // DAY_OF_MONTH
			366, // DAY_OF_YEAR
			SATURDAY, // DAY_OF_WEEK
			6, // DAY_OF_WEEK_IN
			PM, // AM_PM
			11, // HOUR
			23, // HOUR_OF_DAY
			59, // MINUTE
			59, // SECOND
			999, // MILLISECOND
			14 * ONE_HOUR, // ZONE_OFFSET
			2 * ONE_HOUR // DST_OFFSET (double summer time)
	};

	/**
	 * Converts calendar field values to the time value (millisecond offset from
	 * the <a href="Calendar.html#Epoch">Epoch</a>).
	 * 
	 * @throws IllegalArgumentException
	 *             if any calendar fields are invalid.
	 */
	protected void computeTime() {
		int year = fields[Calendar.YEAR];
		int month = fields[Calendar.MONTH] + 1;
		int day = fields[Calendar.DAY_OF_MONTH];
		int hour = fields[Calendar.HOUR_OF_DAY];

		int minute = fields[Calendar.MINUTE];
		int second = fields[Calendar.SECOND];
		int milsec = fields[Calendar.MILLISECOND];

		if (hour < 0 || hour > 24)
			throw new IllegalArgumentException();
		if (year < 0)
			throw new IllegalArgumentException();
		if (month < 1 || month > 12)
			throw new IllegalArgumentException();
		if (day < 1 || day > 31)
			throw new IllegalArgumentException();
		if ((month == 4 || month == 6 || month == 9 || month == 11) && day > 30)
			throw new IllegalArgumentException();
		if (month == 2 && day >= 29)
			throw new IllegalArgumentException();

		if (month <= 2) {
			month = month + 12;
			year = year - 1;
		}

		double wjd2 = 365.0 * (year) + Math.floor(30.6001 * (month + 1)) + day
				- 64;
		time = Math
				.round((((((wjd2 * 24 + hour) * 60) + minute) * 60) + second) * 1000)
				+ milsec;
	}

	/**
	 * Converts the time value (millisecond offset from the <a
	 * href="Calendar.html#Epoch">Epoch</a>) to calendar field values. The time
	 * is <em>not</em> recomputed first; to recompute the time, then the
	 * fields, call the <code>complete</code> method.
	 * 
	 * @see java.util.Calendar#complete
	 */
	protected void computeFields() {
		double wjd = (double) time / (24 * 60 * 60 * 1000.0);
		long z = Math.round(Math.floor(wjd)) + 365;
		long f = time - ((z - 365) * 24 * 60 * 60 * 1000);
		long c = (z - 59) / 365;
		long d = 365 * c;
		long e = Math.round(Math.floor((z - d + 64) / 30.6001));

		int day = (int) (z - d - Math.round(Math.floor(30.6001 * e)) + 64);
		int month;
		if (e < 14) {
			month = (int) e - 1;
		} else {
			month = (int) e - 13;
		}

		int year;
		if (month > 2) {
			year = (int) c - 1;
		} else {
			year = (int) c;
		}

		int hour = (int) Math.floor(f / (1000 * 60 * 60));
		long m = f - hour * 1000 * 60 * 60;
		int minute = (int) Math.floor(m / (1000 * 60));
		long s = m - minute * 1000 * 60;
		int second = (int) Math.floor(s / 1000);
		int milsec = (int) (s - second * 1000);

		fields[Calendar.YEAR] = year;
		fields[Calendar.MONTH] = month - 1;
		fields[Calendar.DAY_OF_MONTH] = day;
		fields[Calendar.HOUR_OF_DAY] = hour;

		if (hour <= 11) {
			fields[Calendar.HOUR] = hour;
			fields[Calendar.AM_PM] = Calendar.AM;
		} else {
			fields[Calendar.HOUR] = hour - 12;
			fields[Calendar.AM_PM] = Calendar.PM;
		}

		fields[Calendar.MINUTE] = minute;
		fields[Calendar.SECOND] = second;
		fields[Calendar.MILLISECOND] = milsec;
	}

	/**
	 * Adds or subtracts the specified amount of time to the given calendar
	 * field, based on the calendar's rules. For example, to subtract 5 days
	 * from the current time of the calendar, you can achieve it by calling:
	 * <p>
	 * <code>add(Calendar.DAY_OF_MONTH, -5)</code>.
	 * 
	 * @param field
	 *            the calendar field.
	 * @param amount
	 *            the amount of date or time to be added to the field.
	 * @see #roll(int,int)
	 * @see #set(int,int)
	 */
	public void add(int field, int amount) {
		switch (field) {
		case Calendar.YEAR: {
			time = time + (long) 365 * 24 * 60 * 60 * 1000 * amount;
			break;
		}
		case Calendar.MONTH: {
			computeFields();
			for (int i = 0; i < Math.abs(amount); i++) {
				if (amount < 0) {
					if (fields[Calendar.MONTH] == 0) {
						fields[Calendar.MONTH] = 11;
						fields[Calendar.YEAR]--;
					} else {
						fields[Calendar.MONTH]--;
					}
				} else {
					if (fields[Calendar.MONTH] == 11) {
						fields[Calendar.MONTH] = 0;
						fields[Calendar.YEAR]++;
					} else {
						fields[Calendar.MONTH]++;
					}
				}
			}
			computeTime();
			break;
		}
		case Calendar.DAY_OF_MONTH: {
			time = time + (long) 24 * 60 * 60 * 1000 * amount;
			break;
		}
		case Calendar.HOUR_OF_DAY:
		case Calendar.HOUR: {
			time = time + (long) 60 * 60 * 1000 * amount;
			// setTimeInMillis(time + (long) 60 * 60 * 1000 * amount);
			break;
		}
		case Calendar.MINUTE: {
			time = time + (long) 60 * 1000 * amount;
			break;
		}
		case Calendar.SECOND: {
			time = time + (long) 1000 * amount;
			break;
		}
		case Calendar.MILLISECOND: {
			time = time + amount;
			break;
		}
		}
		computeFields();
	}

	public void setTimeInMillis(long millis) {
		time = millis;
		computeFields();
		// super.setTimeInMillis(millis);
	}

	public void roll(int field, boolean up) {
		if (up) {
			add(field, 1);
		} else {
			add(field, -1);
		}
	}

	public int getMinimum(int field) {
		return MIN_VALUES[field];
	}

	public int getMaximum(int field) {
		return MAX_VALUES[field];
	}

	public int getGreatestMinimum(int field) {
		return MIN_VALUES[field];
	}

	public int getLeastMaximum(int field) {
		return LEAST_MAX_VALUES[field];
	}
}
